/* daytime.c (C) Copyright Bill Buckels 2008 */

/*

Time of Day Clock (TOD) C64 Digital Clock demonstration program

There are two TOD clocks on the Commodore 64 (C64). They work identically
and are arranged at their respective memory locations identically.

This program uses one of the TOD clocks on the C64 to demonstrate how to
set and how to read the TOD clock in an Aztec C program. It has been built
under Windows XP in the Aztec C65 cross-development environment for the C64.

This program links with the standard native mode C64NAT.LIB provided by Manx
as well as with my own B64NAT.LIB.

Like any C64 native mode program, this program is loaded from BASIC 2. I have
added startup and shutdown code so this program will exit cleanly to BASIC 2
when done.

Errata: The Tenths Register does not seem to work with this program in the
WinVice Emulator. It could be that I am not setting or reading this register
correctly or that I have missed a step but I suspect not. Since I do not
currently have my real C64 set-up to test this I am unable to rule-out a
WinVice problem.

The TOD clock is a special purpose timer for real-time applications.
TOD consists of a 24-hour (AM/PM) clock with 1/10th second resolution. It
is organized into 4 registers: 10ths of seconds, Seconds, Minutes and
Hours. The AM/PM flag is in the MSB of the Hours register for easy bit
testing. Each register reads out in BCD format to simplify conversion for
driving displays, etc. The clock requires an external 60 Hz or 50 Hz
(programmable) TTL level input on the TOD pin for accurate time-keeping.
In addition to time-keeping, a programmable ALARM is provided for
generating an interrupt at a desired time. The ALARM registers or located
at the same addresses as the corresponding TOD registers. Access to the
ALARM is governed by a Control Register bit. The ALARM is write-only; any
read of a TOD address will read time regardless of the state of the ALARM
access bit.

A specific sequence of events must be followed for proper setting and
reading of TOD. TOD is automatically stopped whenever a write to the
Hours register occurs. The clock will not start again until after a write
to the 10ths of seconds register. This assures TOD will always start at
the desired time. Since a carry from one stage to the next can occur at
any time with respect to a read operation, a latching function is
included to keep all Time Of Day information constant during a read
sequence. All four TOD registers latch on a read of Hours and remain
latched until after a read of 10ths of seconds. The TOD clock continues
to count when the output registers are latched. If only one register is
to be read, there is no carry problem and the register can be read "on
the fly," provided that any read of Hours is followed by a read of 10ths
of seconds to disable the latching.

$DC0F
56335    CIA Control Register B Write bit:
7  Set Alarm/TOD-Clock: 1 = Alarm, 0 = Clock

$DC08
56328 Time of Day, tenth seconds (in BCD). Values: $00-$09.
Read: Current TOD value.
Write: Set TOD or alarm time.

$DC09
56329 Time of Day, seconds (in BCD). Values: $00-$59.
Read: Current TOD value.
Write: Set TOD or alarm time.

$DC0A
56330 Time of Day, minutes (in BCD). Values: $00-$59.
Read: Current TOD value.
Write: Set TOD or alarm time.

$DC0B
56331 Time of Day, hours (in BCD). Read bits:

Bits #0-#5: Hours.

Bit #7: 0 = AM; 1 = PM.

Write: Set TOD or alarm time.


*/

#define SETCLOCK 56335

#define THOUR 56331
#define TMIN  56330
#define TSEC  56329
#define TENTH 56328


/* macros for converting to and from C64 BCD format */

#define atobcd(x) (((x/10)<<4)|(x-((x/10)*10)))
#define bcdtoa(x) (((x>>4)*10) + (x&0xf))

#define CMAIN 1

#include <poke.h>
#include <colors.h>


char buf[40];

main()
{
    int hours, minutes, seconds, tenths;
    int ihours, iminutes, iseconds, itenths;
    int h, m, s, t, pm, ch;
    register ireg;

    scr_clear(); /* clear screen */
    scr_txtcolor(C64_BWHITE);   /* bwhite text */

    /* clear all keypresses on entry */
    while (kbhit());
    /* clear the buffer */
    kbflush();

    /* locate text screen cursor for input */
    scr_curs(12,6);
	printf("HOURS 0-23  : ");gets(buf); hours = atoi(buf);
	scr_curs(13,6);
	printf("MINUTES 0-59: ");gets(buf); minutes = atoi(buf);
	scr_curs(14,6);
	printf("SECONDS 0-59: ");gets(buf); seconds = atoi(buf);

    /* clear all keypresses again to avoid run-on */
    while (kbhit());
    /* clear the buffer */
    kbflush();


	if (hours < 0 || hours > 23)hours = 0;
	if (hours > 11) {
		hours = (hours-12);
		pm = 128; /* set high bit of hours reg if PM */
	}
	else {
	    pm = 0;
	}
	if (hours == 0)hours=12;

	if (minutes < 0 || minutes > 59)minutes = 0;
	if (seconds < 0 || seconds > 59)seconds = 0;
	tenths = 0;

    poke[SETCLOCK] &= 127; /* UNSET TIME OF DAY CLOCK - CRB */

    h = atobcd(hours);
    m = atobcd(minutes);
    s = atobcd(seconds);
    t = tenths;

	poke[THOUR] = h + pm; /* add high bit if PM */
	poke[TMIN] =  m;
	poke[TSEC] =  s;
	poke[TENTH] = t;

    scr_clear(); /* clear screen */


    scr_curs(12,6);
    if (pm == 0)
       printf("%02d:%02d:%02d.%02d AM\n",hours,minutes,seconds,tenths);
    else
       printf("%02d:%02d:%02d.%02d PM\n",hours,minutes,seconds,tenths);

    while (!kbhit()) {

		h = peek[THOUR];
		m = peek[TMIN];
		s = peek[TSEC];
		t = peek[TENTH]; /* reading of tenths does not seem
		                    to work in WinVice */

		pm = ((h >> 7) <<7);
		h = (h - pm);

		ihours =  bcdtoa(h);
		iminutes = bcdtoa(m);
		iseconds = bcdtoa(s);
		itenths = bcdtoa(t);

		if (iseconds != seconds||iminutes!=minutes||ihours!=hours) {

			hours = ihours;
			minutes = iminutes;
			seconds = iseconds;
			tenths = itenths;

            scr_curs(12,6);
			if (pm == 0)
			   printf("%02d:%02d:%02d.%02d AM\n",hours,minutes,seconds,tenths);
			else
			   printf("%02d:%02d:%02d.%02d PM\n",hours,minutes,seconds,tenths);
		}
	}
    /* clear all keypresses on exit */
    while (kbhit()); /* wait for release */
    getchar();
    kbflush();
    scr_reset(); /* clear again and reset to default and UPPER case */
    exit(0);
}
