/* jiffy.c (C) Copyright Bill Buckels 2008 */

/*

Jiffy Clock C64 interval timer demonstration program

This Aztec C program uses the Commodore 64's interval timer
("jiffy clock") to demonstrate a delay function with a reasonably
high resolution. It has been built under Windows XP in the Aztec C65
cross-development environment for the C64.

This program links with the standard native mode C64NAT.LIB provided
by Manx. Like any C64 native mode program, this program is loaded
from BASIC 2. I have added startup and shutdown code so this program
will exit cleanly to BASIC 2 when done.

The jiffy clock which is referenced by the BASIC variables TI and TI$ is
different from the two Time of Day (TOD) clocks on the C64. Its value is
set at zero (initialized) when you power-up the system and advances in
1/60 second intervals from that point on. It can also be assigned an
arbitrary starting point within its range of 0-518399 similar to the way
you set your wristwatch.

The jiffy clock is updated by a carefully timed sequence of instructions
in the interrupt service routine of the C64. However if the interrupt
routine is suspended (as, for example, during cassette loads or saves),
the jiffy clock loses its accuracy. It is for reasons of this accuracy loss
that at least one C64 program (a game called "Centipede") also used the
the second TOD clock (which has a lower resolution but does not lose
accuracy) to resynch the jiffy clock.

In this program, I have taken advantage of the jiffy clock's ability
to be cleared to 0. I then read the first two bytes of the jiffy clock
as an integer in a tight loop in my delay function to advance toward
a target value.

I have kept this fairly simple, and written a general purpose timeout
function that is suitable for timing of simple tasks.

I did not attempt anything as elaborate as synching my DELTA time in
my delay with the ALPHA in the TOD clock. I have also tried where
I considered prudent to stay within the range of integer arithmetic
avoiding floating point math (and long integers except when scaling
from non-native units to 60ths of a second).

*/

#define SETCLOCK 56335

#define THOUR 56331
#define TMIN  56330
#define TSEC  56329
#define TENTH 56328

#define peek poke

/* integer to c64 binary coded decimal */
#define atobcd(x) (((x/10)<<4)|(x-((x/10)*10)))
#define bcdtoa(x) (((x>>4)*10) + (x&0xf))

/* keyboard pointers */
char *key = (char *)203;
char *kbuf = (char *)631;
char *klen = (char *)198;

/* if you are going to poke around
   this is the best way that I have found */
char *poke = (char *)0;  /* a zero pointer */

char buf[40];

/* scaling factors native, ms. 10ths, 18.2 ticks (ibm) */
/* used by delay function */
#define NSCALE 60
#define ISCALE 182
#define HSCALE 100
#define MSCALE 1000

/*

$00A0-$00A2
160-162 Value of TI variable, time of day, increased by 1 every 1/60
second (on PAL machines).
Values: $000000-$4F19FF, 0-518399 (on PAL machines).

*/

/* just use the first two bytes - stick to integers for efficency */
/* for smaller values a char pointer would be just fine */
int *jiffy = (int *)160;

delay(duration, scale)
int duration, scale;
{

	int cnt = 0;
	long lduration; /* accumulator to avoid overflow */

	if (duration < 1)return;

    /* re-initialize the jiffy clock */
	/* zero seed value to avoid resolution loss */
    poke[160] = 0;
    poke[161] = 0;
    peek[162] = 0;

    /* scale the duration based on the duration units */
	/* simple ratio over proportion using whole numbers */
	/* so multiply first then divide */
	switch(scale)
	{
		             /* milliseconds */
	    case MSCALE: /* need to convert to 60ths from 1000ths */
                     /* 100ths */
                     /* need to convert to 60ths from 100ths */
        case HSCALE: lduration = (long)duration;
                     duration = (int)((lduration * 60)/scale);
                     break;

                     /* IBM scale */
		case ISCALE: /* need to convert to 60ths from 18.2 ticks */
		           lduration = (long)duration;
		           duration = (int)((lduration * 600)/182);
		           break;

		case NSCALE: /* native scale so just use as is */
		default:   break;

	}

	/* continue to check TI until done */
	/* tight loop */
	while (cnt < duration) cnt = jiffy[0];


}



main()
{
    int hours=0, minutes=0, seconds=0, tenths=0;
    int h, m, s, t, pm, duration;
    register ireg;

    putchar(147); /* clear screen */
    putchar(5);   /* bwhite text */

    /* clear all keypresses on entry */
    while (key[0] != 0x40);
    /* clear the buffer */
    for (ireg = 9; ireg > -1; ireg--) kbuf[ireg]=0;
    klen[0] = 0;

	do {
		printf("ENTER DELAY IN 60THS OF A SECOND : ");
		gets(buf);
		duration = atoi(buf);
	}while (duration < 1);

    printf("\nDELAY IN SECONDS (APPROXIMATE)\n%02d:%02d:%02d START\n",hours,minutes,seconds);


/* The time of day clock is not required to use the delay function.
	However in this program it results in a reasonably effective cross-check
	for timer resolution which by the way may not be terribly accurate on very
	small values due to the overhead of the setup instructions required by the
	delay function. */

    jiffy[0]=0;
    ireg = duration; /* load register */

	poke[SETCLOCK] &= 127; /* UNSET TIME OF DAY CLOCK - CRB */
	poke[THOUR] = 0;
	poke[TMIN] =  0;
	poke[TSEC] =  0;
	poke[TENTH] = 0;


    /* in a production version if small delays were absolutely
       necessary we probably wouldn't want to bother with making a
       function call and certainly not to incur the extra overhead
       of start and end times...

       So I have commented out the function call below which
       works fine but uses extra time saving and restoring the
       stack and so forth, and just do a timing loop.

    */

    /* do it */
    /* delay(duration); */

    while (*jiffy < ireg);

	h = peek[THOUR]; /* unlatch */
	m = peek[TMIN];
	s = peek[TSEC];
	t = peek[TENTH]; /* latch */

    /* clear all keypresses again to avoid run-on */
    while (key[0] != 0x40);
    /* clear the buffer */
    for (ireg = 9; ireg > -1; ireg--) kbuf[ireg]=0;
    klen[0] = 0;

	pm = ((h >> 7) <<7);
	h = (h - pm);

    hours =  bcdtoa(h);
    minutes = bcdtoa(m);
	seconds = bcdtoa(s);

	printf("%02d:%02d:%02d STOP\n",hours,minutes,seconds);

    /* clear all keypresses on exit */
    while (key[0] != 0x40); /* wait for release */
    /* clear the buffer */
    for (ireg = 9; ireg > 1; ireg--) kbuf[ireg]=0;
    kbuf[1] = 13; /* stuff the buffer - enter key */
    kbuf[0] = 154;/* change text back to light blue */
    klen[0] = 2;
    getchar();
    exit(0);
}
