/* snd2c64 (C) Copyright Bill Buckels 2008 */


/* converts binary musical files of type .SND   */
/* which are IBM files containing frequency, duration pairs in IBM format */
/* to character arrays for embedding in Commodore 64 C code
   additionally creates a binary file of the same format for optional
   playing from disk.
*/

/* the output file is midi note number, duration in 18.2 ticks per second  */
/* the use of these need not be confined to the C64 and although they are
   fairly primitive bitmusic they can be enhanced to play reasonably
   well in other programs */

/* on the C64 I define my own ADSR sound waves to a variety
  of simple intruments which are presets and then convert
  the midi note that this program resolves to a C64 frequency in a
  xref table.

  I just use a relative timing loop for duration which has
  turned-out to be reasonably accurate in translating from 18.2 tps.
  Since these SND files are time based rather than tempo based
  it is trivial to translate these to a midi note on, midi note off
  in a more sophisticated platform.


  I have also defined my own C64 fequency table. The table in the
  C64 user's guide is 100% inaccurate as near as I can tell.

  */



/*
IBM format (similar to GWBASIC sound statement)

frequency    The frequency of the sound in hertz; a value in the range
			   37 through 32,767.
duration     The number of system clock ticks the sound lasts; a value
			   in the range 0 through 65,535. There are 18.2 clock ticks
			   per second.

*/

/*

Octave # Note Numbers

C C# D D# E F F# G G# A A# B
-1 0 1 2 3 4 5 6 7 8 9 10 11
0 12 13 14  15 16 17 18 19 20 21 22  23
1 24 25 26 27 28 29 30 31 32 33 34 35
2 36 37 38 39 40 41 42 43 44 45 46 47
3 48 49 50 51 52 53 54  55 56 57 58 59
4 60 61 62 63 64 65 66 67 68 69 70 71
5 72 73 74 75 76 77 78 79 80 81 82 83
6 84 85 86 87 88 89 90 91 92 93 94 95
7 96 97 98 99 100 101 102 103 104 105 106 107
8 108 109 110 111 112 113 114 115 116 117 118 119
9 120 121 122 123 124 125  126 127


Note: The MIDI specification only defines note number 60 as "Middle C",
and all other notes are relative. The absolute octave number designations
shown here are based on Middle C = C4, which is an arbitrary assignment.


*/

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>


/* note although i do not use this table in this program
   i left it here for reference */
/* the following C64 SID frequencies represent
   general midi octaves -1 through 6 */
/* to find midi note 60 (mid C) the offset is 60 */

unsigned int cfreq[95]={
  274,  291,  308,  326,  345,  366,  388,  411,  435,  461,  488,  517,
  549,  582,  616,  653,  691,  733,  776,  822,  871,  923,  977, 1035,
 1099, 1164, 1233, 1306, 1383, 1466, 1553, 1645, 1742, 1846, 1955, 2071,
 2198, 2328, 2466, 2613, 2767, 2932, 3106, 3290, 3485, 3692, 3911, 4143,
 4396, 4657, 4933, 5226, 5535, 5864, 6212, 6580, 6970, 7384, 7822, 8287,
 8793, 9315, 9867,10452,11071,11728,12424,13161,13941,14769,15645,16574,
17557,18600,19704,20874,22113,23426,24818,26292,27853,29508,31261,33118,
35084,37170,39378,41718,44196,46822,49606,52554,55676,58986,62492};


/*

Relationships of Notes and Frequencies on the IBM-PC

Note Frequency
C 130.810
D 146.830
E 164.810
F 174.610
G 196.000
A 220.000
B 246.940
C 261.630
D 293.660
E 329.630
F 349.230
G 392.000
A 440.000
B 493.880
C 523.250 middle C (midi note #60)
D 587.330
E 659.260
F 698.460
G 783.990
A 880.000
B 987.770
C 1046.500
D 1174.700
E 1318.500
F 1396.900
G 1568.000
A 1760.000
B 1975.500

*/


/* ------------------------------------------------------------------- */
/* Translates a frequency within a range to a corresponding midi note. */
/* ------------------------------------------------------------------- */
int xplay(int xfrequency)
{

  /* octaves out of C64 range are ignored */
  if(xfrequency < 32 || xfrequency > 8192 )
  {
    /* if we are out of range don't play anything */
    return -1;
  }

  /* octave 0 */
  if(xfrequency<63)
  {
  if(xfrequency<34 ) return  12;
  if(xfrequency<36 ) return  13;
  if(xfrequency<38 ) return  14;
  if(xfrequency<40 ) return  15;
  if(xfrequency<42 ) return  16;
  if(xfrequency<45 ) return  17;
  if(xfrequency<48 ) return  18;
  if(xfrequency<51)  return  19;
  if(xfrequency<54)  return  20;
  if(xfrequency<57)  return  21;
  if(xfrequency<60)  return  22;
                     return  23;
  }

  /* first octave */
  if(xfrequency<127)
  {
  if(xfrequency<67 ) return  24;
  if(xfrequency<71 ) return  25;
  if(xfrequency<76 ) return  26;
  if(xfrequency<80 ) return  27;
  if(xfrequency<85 ) return  28;
  if(xfrequency<90 ) return  29;
  if(xfrequency<96 ) return  30;
  if(xfrequency<101) return  31;
  if(xfrequency<107) return  32;
  if(xfrequency<114) return  33;
  if(xfrequency<120) return  34;
                     return  35;
  }

  /* second octave */
  if(xfrequency<255)
  {
  if(xfrequency<136) return  36;
  if(xfrequency<143) return  37;
  if(xfrequency<152) return  38;
  if(xfrequency<160) return  39;
  if(xfrequency<170) return  40;
  if(xfrequency<180) return  41;
  if(xfrequency<190) return  42;
  if(xfrequency<202) return  43;
  if(xfrequency<215) return  44;
  if(xfrequency<227) return  45;
  if(xfrequency<240) return  46;
                     return  47;
  }

  /* third octave */
  if(xfrequency<512)
  {
   if(xfrequency<270)return  48;
   if(xfrequency<286)return  49;
   if(xfrequency<304)return  50;
   if(xfrequency<322)return  51;
   if(xfrequency<340)return  52;
   if(xfrequency<360)return  53;
   if(xfrequency<382)return  54;
   if(xfrequency<406)return  55;
   if(xfrequency<428)return  56;
   if(xfrequency<454)return  57;
   if(xfrequency<480)return  58;
                     return  59;
   }

  /* fourth octave */
  if(xfrequency<1024)
  {
  if(xfrequency<540)return 60; /* middle C (C4) */
  if(xfrequency<572)return 61;
  if(xfrequency<605)return 62;
  if(xfrequency<643)return 63;
  if(xfrequency<680)return 64;
  if(xfrequency<720)return 65;
  if(xfrequency<765)return 66;
  if(xfrequency<810)return 67;
  if(xfrequency<855)return 68;
  if(xfrequency<910)return 69;
  if(xfrequency<965)return 70;
                    return 71;
  }

  /* fifth octave */
  if(xfrequency<2048)
  {
  if(xfrequency<1080) return 72;
  if(xfrequency<1144) return 73;
  if(xfrequency<1210) return 74;
  if(xfrequency<1286) return 75;
  if(xfrequency<1340) return 76;
  if(xfrequency<1440) return 77;
  if(xfrequency<1530) return 78;
  if(xfrequency<1620) return 79;
  if(xfrequency<1710) return 80;
  if(xfrequency<1820) return 81;
  if(xfrequency<1930) return 82;
                      return 83;
  }

  if (xfrequency < 4096)
  {

   /* sixth octave */
   /* seventh, eight, and ninth octave ignored */
   if(xfrequency<2160)return 84;
   if(xfrequency<2288)return 85;
   if(xfrequency<2420)return 86;
   if(xfrequency<2572)return 87;
   if(xfrequency<2680)return 88;
   if(xfrequency<2880)return 89;
   if(xfrequency<3060)return 90;
   if(xfrequency<3240)return 91;
   if(xfrequency<3420)return 92;
   if(xfrequency<3640)return 93;
   if(xfrequency<3860)return 94;
                      return 95;
  }

   /* seventh octave */
   /* eight, and ninth octave ignored */
   if(xfrequency<4320)return 96;
   if(xfrequency<4576)return 97;
   if(xfrequency<4840)return 98;
   if(xfrequency<5144)return 99;
   if(xfrequency<5360)return 100;
   if(xfrequency<5760)return 101;
   if(xfrequency<6120)return 102;
   if(xfrequency<6480)return 103;
   if(xfrequency<6840)return 104;
   if(xfrequency<7280)return 105;
   if(xfrequency<7720)return 106;
   return 107;
}

main(int argc, char *argv[])
{
   FILE *fp,*fp2,*fp3;
   char *wordptr;
   unsigned freq, dur;

   int frequency,duration,counter=0;
   char scratchbuffer[128],writebuffer[128], binbuffer[128];

   if(argc!=2)
   {
    printf("SND filename required...");
    exit(0);
    }

   if((fp=fopen(argv[1],"rb"))==NULL)
   {
    perror(argv[1]);
    exit(0);
    }

   strcpy(scratchbuffer,argv[1]);
   wordptr=strtok(scratchbuffer,".");
   sprintf(writebuffer,"%s.TXT",scratchbuffer);
   sprintf(binbuffer,"%s.SIB",scratchbuffer);
   fp2=fopen(writebuffer,"w");
   fp3=fopen(binbuffer,"wb");

   fprintf(fp2,
   "/* musical array created from IBM snd file %s */\n",argv[1]);
   fprintf(fp2,
   "/* array structure is midi note number, duration (18.2 ticks sec) */\n");
   fprintf(fp2,
   "char %s[]={\n",scratchbuffer);

    while((frequency=getw(fp))!=-1){
         duration=fgetc(fp);
         if (duration == 0) continue; /* ignore all notes of 0 duration */

         dur = duration;

         if((frequency=xplay(frequency))<0)
         {
            freq = 255; /* silence */
         }
         else
         {

		    freq = frequency;
         }
         fprintf(fp2,"%3d, %3d,",freq,dur);

         fputc(freq,fp3);
         fputc(dur,fp3);

         counter++;
         if(counter==8){         /* wide enough */
            fprintf(fp2,"\n");
            counter=0;
         }
    }
  	 /* use 0 as the terminator */
  	fprintf(fp2,"%3d, %3d};\n",0,0);


    fputc(0,fp3);
    fputc(0,fp3);

   fclose(fp);
   fclose(fp2);
   fclose(fp3);
   exit(0);

}